# SPDX-FileCopyrightText: 2024-present Davis Vann Bennett <davis.v.bennett@gmail.com>
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import numpy as np
from yarl import URL
import zarr
import polars as pl
from neuroglancer import ImageLayer, AnnotationLayer, ViewerState, CoordinateSpace
from matchviz.bigstitcher import image_name_to_tile_coord
from matchviz.core import get_url, parse_url
from matchviz.core import get_histogram_bounds_batch
from matchviz.core import tile_coordinate_to_rgba
from matchviz.neuroglancer_styles import NeuroglancerViewerStyle
from functools import reduce
from concurrent.futures import ThreadPoolExecutor
from matplotlib import colors
import matplotlib.pyplot as plt
from xarray_ome_ngff import read_multiscale_group
import structlog

pool = ThreadPoolExecutor(max_workers=10)


# we assume here that there's no need to parametrize t
def plot_points(points_df: pl.DataFrame, image_group_path: str):
    images_xarray = read_multiscale_group(
        zarr.open_group(image_group_path, mode="r"),
        array_wrapper={"name": "dask_array", "config": {"chunks": "auto"}},
    )

    loc_xyz = np.array(points_df["loc_xyz"].to_list())

    fig, axs = plt.subplots(ncols=2, nrows=2, dpi=200, figsize=(8, 8))

    dims = ("x", "y", "z")
    img = images_xarray["4"].drop_vars(("t", "c")).squeeze()
    pairs = ("z", "y"), (), ("z", "x"), ("x", "y")

    for idx, pair in enumerate(pairs):
        if pair != ():
            plot_x, plot_y = sorted(pair)
            axis = axs.ravel()[idx]
            proj_dim = tuple(set(dims) - set(pair))[0]
            proj = img.max((proj_dim)).compute()
            proj.name = f"proj_{proj_dim}"
            proj.plot.imshow(
                x=plot_x,
                y=plot_y,
                ax=axis,
                robust=True,
                norm=colors.LogNorm(),
                cmap="gray_r",
            )

            axis.scatter(
                loc_xyz[:, dims.index(plot_x)],
                loc_xyz[:, dims.index(plot_y)],
                marker="o",
                facecolor="y",
                edgecolor="y",
                alpha=0.1,
            )

            axis.set_xlabel(plot_x)
            axis.set_ylabel(plot_y)

    fig.savefig("foo.svg")


def create_neuroglancer_state(
    image_url: str | URL,
    points_url: str | URL | None,
    matches_url: str | URL | None,
    style: NeuroglancerViewerStyle,
):
    if points_url is not None:
        points_url_parsed = parse_url(points_url)
    else:
        points_url_parsed = None

    if matches_url is not None:
        matches_url_parsed = parse_url(matches_url)
    else:
        matches_url_parsed = None

    image_url_parsed = parse_url(image_url)
    log = structlog.get_logger()

    image_group = zarr.open_group(store=str(image_url_parsed), path="", mode="r")
    image_sources = {}
    points_sources = {}
    matches_sources = {}
    space = CoordinateSpace(
        names=["z", "y", "x"],
        scales=[
            100,
        ]
        * 3,
        units=[
            "nm",
        ]
        * 3,
    )

    state = ViewerState(
        dimensions=space, cross_section_scale=1000, projection_scale=500_000
    )

    # read the smallest images in the pyramid
    subgroups = tuple(image_group.groups())
    subgroup_urls = tuple(str(get_url(g)) for _, g in subgroups)
    bounds = get_histogram_bounds_batch(subgroup_urls, pool)
    histogram_bounds = {k: v for k, v in zip(image_group.group_keys(), bounds)}

    for fname, sub_group in subgroups:
        subgroup_url = get_url(sub_group)
        image_sources[fname] = f"zarr://{subgroup_url}"

        annotation_dir = fname.removesuffix(".zarr") + ".precomputed"

        if points_url is not None:
            point_url = points_url_parsed.joinpath(annotation_dir)
            points_sources[fname] = f"precomputed://{point_url}"

        if matches_url is not None:
            match_url = matches_url_parsed.joinpath(annotation_dir)
            matches_sources[fname] = f"precomputed://{match_url}"

    # bias the histogram towards the brighter values
    hist_min, hist_max = reduce(
        lambda old, new: (max(old[0], new[0]), max(old[1], new[1])),
        histogram_bounds.values(),
    )
    log.info(f"Using histogram bounded between ({hist_min}, {hist_max})")
    window_min = int(hist_min) - abs(int(hist_min) - int(hist_max)) // 3
    if window_min < 0:
        window_min = 0
    window_max = int(hist_max) + abs(int(hist_min) - int(hist_max)) // 3
    image_shader_controls = {
        "normalized": {
            "range": [int(hist_min), int(hist_max)],
            "window": [window_min, window_max],
        }
    }

    annotation_shader = r"void main(){setColor(prop_point_color());}"

    if style == "images_split":
        for fname, im_source in image_sources.items():
            coordinate = image_name_to_tile_coord(fname)
            name_base = f"x={coordinate['x']}, y={coordinate['y']}, z={coordinate['z']}, ch={coordinate['ch']}"
            color = tile_coordinate_to_rgba(coordinate)
            color_str = "#{0:02x}{1:02x}{2:02x}".format(*color)
            shader = (
                "#uicontrol invlerp normalized()\n"
                f'#uicontrol vec3 color color(default="{color_str}")\n'
                "void main(){{emitRGB(color * normalized());}}"
            )

            state.layers.append(
                name=f"{name_base}/img",
                layer=ImageLayer(
                    source=im_source,
                    shaderControls=image_shader_controls,
                    shader=shader,
                ),
            )
            if points_url is not None:
                point_source = points_sources[fname]
                state.layers.append(
                    name=f"{name_base}/points",
                    layer=AnnotationLayer(
                        source=point_source, shader=annotation_shader
                    ),
                )
            if matches_url is not None:
                match_source = matches_sources[fname]
                state.layers.append(
                    name=f"{name_base}/matches",
                    layer=AnnotationLayer(
                        source=match_source, shader=annotation_shader
                    ),
                )

    elif style == "images_combined":
        state.layers.append(
            name="images",
            layer=ImageLayer(
                source=list(image_sources.values()),
                shader_controls=image_shader_controls,
            ),
        )
        if points_url is not None:
            state.layers.append(
                name="points",
                layer=AnnotationLayer(
                    source=list(points_sources.values()), shader=annotation_shader
                ),
            )
        if matches_url is not None:
            state.layers.append(
                name="matches",
                layer=AnnotationLayer(
                    source=list(matches_sources.values()), shader=annotation_shader
                ),
            )
    else:
        msg = f"Style {style} not recognized. Style must be one of 'images_combined' or 'images_split'"
        raise ValueError(msg)
    return state



from __future__ import annotations
import numpy as np
import polars as pl
from pydantic_ome_ngff.v04.multiscale import MultiscaleMetadata
import zarr
from yarl import URL
import pathlib

import os
from typing import Literal, Sequence

from matchviz.types import Coords, TileCoordinate


def get_url(node: zarr.Group | zarr.Array) -> URL:
    """
    Get a URL from a zarr array or group pointing to its location in storage
    """
    store = node.store
    if hasattr(store, "path"):
        if hasattr(store, "fs"):
            if isinstance(store.fs.protocol, Sequence):
                protocol = store.fs.protocol[0]
            else:
                protocol = store.fs.protocol
        else:
            protocol = "file"

        # fsstore keeps the protocol in the path, but not s3store
        if "://" in store.path:
            store_path = store.path.split("://")[-1]
        else:
            store_path = store.path
        return URL(f"{protocol}://{os.path.join(store_path, node.path)}")
    else:
        msg = (
            f"The store associated with this object has type {type(store)}, which "
            "cannot be resolved to a url"
        )
        raise ValueError(msg)


def ome_ngff_to_coords(url: str | URL) -> Coords:
    multi_meta = MultiscaleMetadata(
        **zarr.open_group(str(url), mode="r").attrs.asdict()["multiscales"][0]
    )
    scale = multi_meta.datasets[0].coordinateTransformations[0].scale
    trans = multi_meta.datasets[0].coordinateTransformations[1].translation
    return {
        axis.name: {"scale": s, "trans": t}
        for axis, s, t in zip(multi_meta.axes, scale, trans)
    }  # type: ignore


def scale_point(point: Sequence[float], params: Sequence[float]):
    return np.multiply(point, params)


def scale_points(points_df: pl.DataFrame, coords: Coords):
    col = "loc_xyz"
    dims = ("x", "y", "z")
    col_index = points_df.columns.index(col)

    scale = [coords[dim]["scale"] for dim in dims]  # type: ignore
    new_col = scale_point(points_df[col].to_list(), scale)

    return points_df.clone().replace_column(
        col_index, pl.Series(name=col, values=new_col.tolist())
    )


def translate_point(point: Sequence[float], params: Sequence[float]):
    return np.add(point, params)


def translate_points(points_df: pl.DataFrame, coords: Coords):
    """
    Apply a translation
    """
    col = "loc_xyz"
    dims: tuple[Literal["x", "y", "z"], ...] = ("x", "y", "z")
    col_index = points_df.columns.index(col)
    local_scale = [coords[dim]["scale"] for dim in dims]
    local_trans = [coords[dim]["trans"] for dim in dims]

    trans = np.divide(local_trans, local_scale)  # type: ignore
    new_col = translate_point(points_df[col].to_list(), trans)

    return points_df.clone().replace_column(
        col_index, pl.Series(name=col, values=new_col.tolist())
    )


def transform_points(coords: Coords, points: np.ndarray):
    translate_points(coords, points)
    scale_points(coords, points)


def get_percentiles(image_url: str) -> tuple[int, int]:
    """
    Get the 5th and 95th percentiles from of the smallest array in a multiscale group
    """
    group = zarr.open_group(image_url, mode="r")
    arrays_sorted = sorted(group.arrays(), key=lambda kv: np.prod(kv[1].shape))
    smallest = arrays_sorted[0][1][:]
    return np.percentile(smallest, (5, 95))


def get_histogram_bounds_batch(group_urls: tuple[str, ...], pool):
    return tuple(pool.map(get_percentiles, group_urls))


def tile_coordinate_to_rgba(coord: TileCoordinate) -> tuple[int, int, int, int]:
    """
    generate an RGBA value from a tile coordinate. This ensures that adjacent tiles have different
    colors. It's not a nice lookup table by any measure.
    """
    mod_map = {}
    for key in ("x", "y", "z"):
        mod_map[key] = coord[key] % 2  # type: ignore
    lut = {
        (0, 0, 0): ((255, 0, 0, 255)),
        (1, 0, 0): ((0, 255, 0, 255)),
        (0, 1, 0): ((0, 0, 255, 255)),
        (1, 1, 0): ((255, 255, 0, 255)),
        (0, 0, 1): ((0, 255, 255, 255)),
        (1, 0, 1): ((191, 191, 191, 255)),
        (0, 1, 1): ((0, 128, 128, 255)),
        (1, 1, 1): ((128, 128, 0, 255)),
    }

    return lut[tuple(mod_map.values())]


def tokenize(data: Sequence[float]) -> Sequence[int]:
    uniques = sorted(np.unique(data).tolist())
    return [uniques.index(d) for d in data]


def tokenize_tile_coords(tile_coords: dict[int, Coords]) -> tuple[tuple[int, int], ...]:
    """
    Convert positions in world coordinates to positions in a space where the only coordinates
    allowed are tuples of integers.
    """

    tile_positions = {k: (v["x"], v["y"], v["z"]) for k, v in tile_coords.items()}
    tile_x, tile_y, tile_z = zip(*tuple(t for t in tile_positions.values()))
    tile_coords_normed = tuple(zip(*map(tokenize, (tile_x, tile_y, tile_z))))
    return tile_coords_normed


def parse_url(data: object) -> URL:
    """
    Parse input as a URL
    """
    if isinstance(data, URL):
        return data
    if isinstance(data, str):
        maybe_url = URL(data.rstrip("/"))
        if maybe_url.scheme == "":
            if maybe_url.is_absolute():
                return maybe_url.with_scheme("file")
            else:
                return URL.build(scheme="file", path=maybe_url.path.rstrip("/"))
        else:
            return maybe_url
    if isinstance(data, pathlib.Path):
        return URL.build(scheme="file", path=str(data))
    else:
        raise TypeError(
            f"Invalid input. Expected a str, URL, or pathlib.path. Got {type(data)}"
        )



import matplotlib.pyplot as plt
import polars as pl
import numpy as np


def plot_matches(
    *, data: pl.DataFrame, dataset_name: str, invert_x, invert_y
) -> plt.Figure:
    fig_w = 8
    fig_h = 8

    fig, axs = plt.subplots(figsize=(fig_w, fig_h))

    if invert_y:
        axs.invert_yaxis()
    if invert_x:
        axs.invert_xaxis()

    axs.spines[["right", "top"]].set_visible(False)
    axs.grid("on", alpha=0.5)
    completed_pairs = set()
    completed_points = set()
    axs.set_xlabel("Image x coordinate (nm)")
    axs.set_ylabel("Image y coordinate (nm)")

    for row in data.rows():
        row_model = dict(zip(data.schema, row))
        id_self = row_model["id_self"]
        id_other = row_model["id_other"]
        name_self = row_model["image_name"]
        coords_self = row_model["x_coord_self"], row_model["y_coord_self"]

        if name_self not in completed_points:
            axs.scatter(*coords_self, label=name_self, marker=f"${id_self}$", s=200)
            completed_points.add(name_self)

        if (id_self, id_other) not in completed_pairs:
            rows_other = data.filter(pl.col("id_self") == row_model["id_other"])
            coords_other = rows_other.select(
                pl.first("x_coord_self"), pl.first("y_coord_self")
            ).to_numpy()[0]

            # ensure that we don't display symmetric pairs
            completed_pairs.add((id_self, id_other))
            completed_pairs.add((id_other, id_self))

            line_x = np.array([coords_self[0], coords_other[0]])
            line_y = np.array([coords_self[1], coords_other[1]])
            axs.plot(line_x, line_y, color=(0.75, 0.75, 0.75), zorder=0)

            axs.text(
                line_x[0] * 0.65 + line_x[1] * 0.35,
                line_y[0] * 0.65 + line_y[1] * 0.35,
                row_model["num_matches"],
                horizontalalignment="center",
                verticalalignment="top",
                rotation_mode="anchor",
            )
    axs.set_title(f"Number of matches found across tiles in {dataset_name}", wrap=True)
    fig.legend(loc=8, mode="expand", ncols=2, markerscale=0.5)
    # fig.tight_layout()
    plt.subplots_adjust(bottom=0.3)
    return fig



from __future__ import annotations
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
import re
import time
from typing import Annotated, cast, TYPE_CHECKING
from typing_extensions import TypedDict, deprecated

from pydantic import BaseModel, BeforeValidator, Field
from pydantic_zarr.v2 import ArraySpec, GroupSpec

from matchviz.annotation import write_line_annotations, write_point_annotations
from matchviz.core import (
    get_url,
    parse_url,
    tile_coordinate_to_rgba,
    translate_points,
)
from matchviz.types import Coords, TileCoordinate

if TYPE_CHECKING:
    pass
import zarr
import structlog
import fsspec
import numpy as np
import polars as pl
from pydantic_bigstitcher import SpimData2, ViewSetup
from yarl import URL
import neuroglancer


def read_bigstitcher_xml(url: URL, anon: bool = True) -> SpimData2:
    fs, path = fsspec.url_to_fs(str(url), anon=anon)
    bs_xml = fs.cat_file(path)
    bs_model = SpimData2.from_xml(bs_xml)
    return bs_model


# TODO: use this instead of the raw tuple
@dataclass(frozen=True)
class IdMap:
    id_self: int
    id_other: int
    matches_path: str


def get_tilegroup_url(model: SpimData2) -> URL:
    if hasattr(model.sequence_description.image_loader, "s3bucket"):
        bucket = model.sequence_description.image_loader.s3bucket
        image_root_path = model.sequence_description.image_loader.zarr.path
        return URL.build(scheme="s3", authority=bucket, path=image_root_path)
    else:
        # this will be a relative URL
        return URL(model.sequence_description.image_loader.n5.path)


@deprecated(
    "This functionality is deprecated. Use image metadata or the bigstitcher xml file instead."
)
def image_name_to_tile_coord(image_name: str) -> TileCoordinate:
    coords = {}
    for index_str in ("x", "y", "z", "ch"):
        prefix = f"_{index_str}_"
        matcher = re.compile(f"{prefix}[0-9]*")
        matches = matcher.findall(image_name)
        if len(matches) > 1:
            raise ValueError(f"Too many matches! The string {image_name} is ambiguous.")
        substr = matches[0][len(prefix) :]
        if index_str == "ch":
            coords[index_str] = substr
        else:
            coords[index_str] = int(substr)
    coords_out: TileCoordinate = cast(TileCoordinate, coords)
    return coords_out


def parse_idmap(data: dict[str, int]) -> dict[tuple[int, int, str], int]:
    """
    convert {'0,1,beads': 0} to {(0, 1, "beads"): 0}
    """
    parts = map(lambda k: k.split(","), data.keys())
    # convert first two elements to int, leave the last as str
    parts_normalized = map(lambda v: (int(v[0]), int(v[1]), v[2]), parts)
    return dict(zip(parts_normalized, data.values()))


@deprecated(
    "This functionality is deprecated. Use image metadata or the bigstitcher xml file instead."
)
def tile_coord_to_image_name(coord: TileCoordinate) -> str:
    """
    {'x': 0, 'y': 0, 'ch': 561'} -> "tile_x_0000_y_0002_z_0000_ch_488.zarr/"
    """
    cx = coord["x"]
    cy = coord["y"]
    cz = coord["z"]
    cch = coord["ch"]
    return f"tile_x_{cx:04}_y_{cy:04}_z_{cz:04}_ch_{cch}"


def parse_matches(
    *, name: str, data: np.ndarray, id_map: dict[tuple[int, int, str], int]
):
    """
    Convert a name, match data, and an id mapping to a polars dataframe that contains
    pairwise image matching information.
    """
    data_copy = data.copy()

    # get the self id, might not be robust
    match = re.search(r"viewSetupId_(\d+)", name)
    if match is None:
        raise ValueError(f"Could not infer id_self from {name}")

    id_self = int(match.group(1))

    # map from pair index to image id
    remap = {value: key[1] for key, value in id_map.items()}

    # replace the pair id value with an actual image index reference in the last column
    data_copy[:, -1] = np.vectorize(remap.get)(data[:, -1])

    match_result = pl.DataFrame(
        {
            "point_self": data_copy[:, 0],
            "point_other": data_copy[:, 1],
            "id_self": [id_self] * data_copy.shape[0],
            "id_other": data_copy[:, 2],
        }
    )
    return match_result


def load_points_tile(
    url: str | URL, anon: bool = True
) -> tuple[pl.DataFrame, pl.DataFrame | None]:
    """
    Load interest points and optionally correspondences from a bigstitcher-formatted n5 group as
    polars dataframes for a single tile.
    """
    log = structlog.get_logger()
    url_parsed = str(url)
    store = zarr.N5FSStore(url_parsed, mode="r", anon=anon)
    interest_points_group = zarr.open_group(
        store=store, path="interestpoints", mode="r"
    )

    if "id" not in interest_points_group:
        raise ValueError(
            f"Failed to find expected n5 dataset at {get_url(interest_points_group)}/id"
        )
    if "loc" not in interest_points_group:
        raise ValueError(
            f"Failed to find expected n5 dataset at {get_url(interest_points_group)}/loc"
        )

    correspondences_group = zarr.open_group(
        store=store, path="correspondences", mode="r"
    )

    matches_exist = "data" in correspondences_group
    if not matches_exist:
        log.info(f"No matches found for {url_parsed}.")

    # points are saved as [num_points, [x, y, z]]
    loc = interest_points_group["loc"][:]

    ids = interest_points_group["id"][:]
    ids_list = ids.squeeze().tolist()

    if matches_exist:
        id_map = parse_idmap(correspondences_group.attrs["idMap"])
        matches = np.array(correspondences_group["data"])
        match_result = parse_matches(name=url_parsed, data=matches, id_map=id_map)
    else:
        match_result = None
    return pl.DataFrame({"id": ids_list, "loc_xyz": loc}), match_result


def get_tile_coords(bs_model: SpimData2, nominal_grid=True) -> dict[int, Coords]:
    """
    Get the coordinates of all the tiles referenced in bigstitcher xml data. Returns a dict with int
    keys (id numbers of tiles) and Coords values ()
    """
    # get all transformations for all view_setups
    coords_by_view_setup: dict[int, Coords] = {}
    for vs, vr in zip(
        bs_model.sequence_description.view_setups.view_setups,
        bs_model.view_registrations.elements,
    ):
        transforms = tuple(v.to_transform() for v in vr.view_transforms)
        # note that we reverse the order of the transformations, because bigsticher appends new transformations
        # to the front of the list, but we want to iterate with the first applied transform first.
        translation_array = np.array(
            tuple(tuple(t.transform.translation.values()) for t in reversed(transforms))
        )
        translation_array_summed = np.cumsum(translation_array, axis=0)
        if nominal_grid:
            for row in translation_array_summed:
                if not np.all(row == 0):
                    final_translation = row
                    break
        else:
            final_translation = translation_array_summed[-1]

        coords_by_view_setup[int(vs.ident)] = {
            "z": final_translation[2],
            "y": final_translation[1],
            "x": final_translation[0],
        }
    return coords_by_view_setup


class InterestPointsGroupMeta(BaseModel):
    list_version: str = Field(alias="list version")
    pointcloud: str
    type: str


class CorrespondencesGroupMeta(BaseModel):
    correspondences: str
    idmap: Annotated[dict[tuple[int, int, str], int], BeforeValidator(parse_idmap)]


class InterestPointsMembers(TypedDict):
    """
    id is a num_points X 1 array of integer IDs
    loc is a num_points X ndim array of locations in work coordinates
    """

    id: ArraySpec
    loc: ArraySpec


class PointsGroup(GroupSpec[InterestPointsGroupMeta, InterestPointsMembers]):
    members: InterestPointsMembers
    ...


def save_annotations(
    *,
    image_id: int,
    tile_name: str,
    interest_points: URL | str,
    dest_url: URL | str,
    tile_coords: dict[int, Coords],
):
    """
    Load points and correspondences (matches) between a single tile an all other tiles, and save as neuroglancer
    precomputed annotations.

        e.g. dataset = 'exaSPIM_3163606_2023-11-17_12-54-51'
        alignment_id = 'alignment_2024-01-09_05-00-44'

    N5 is organized according to the structure defined here: https://github.com/PreibischLab/multiview-reconstruction/blob/a566bf4d6d35a7ab00d976a8bf46f1615b34b2d0/src/main/java/net/preibisch/mvrecon/fiji/spimdata/interestpoints/InterestPointsN5.java#L54

    If matches are not point, then just the interest points will be saved.
    """

    log = structlog.get_logger(tile_name=tile_name)
    start = time.time()
    log.info(f"Begin saving annotations for image id {image_id}")
    dest_url_parsed = parse_url(dest_url)
    points_url = dest_url_parsed.joinpath(f"points/{tile_name}.precomputed")
    lines_url = dest_url_parsed.joinpath(f"matches/{tile_name}.precomputed")
    interest_points_url_parsed = parse_url(interest_points)

    log.info(f"Saving points to {points_url}")
    log.info(f"Saving matches to {lines_url}")
    # remove trailing slash
    alignment_store = zarr.N5FSStore(str(interest_points_url_parsed).rstrip("/"))

    base_coords = tile_coords[image_id]

    match_group = zarr.open_group(
        store=alignment_store, path="correspondences", mode="r"
    )

    to_access: tuple[int, ...] = (image_id,)
    id_map_normalized = {}
    # tuple of view_setup ids to load
    points_map: dict[int, pl.DataFrame] = {}
    matches_map: dict[int, None | pl.DataFrame] = {}

    matches_exist = "data" in match_group

    if matches_exist:
        log.info("Found matches.")
        id_map = parse_idmap(match_group.attrs.asdict()["idMap"])
        # the idMap attribute uses 0 instead of the actual setup id for the self in this metadata.
        # normalizing replaces that 0 with the actual setup id.
        id_map_normalized = {
            (image_id, *key[1:]): value for key, value in id_map.items()
        }
    else:
        log.info("No matches found.")
    for key in id_map_normalized:
        to_access += (key[1],)

    for img_id in to_access:
        new_name = f"tpId_0_viewSetupId_{img_id}"
        *keep_pre, replace, keep_post = interest_points_url_parsed.parts
        new_url = URL("/".join([*keep_pre, new_name, keep_post]))
        coords = tile_coords[img_id]
        points_data, match_data = load_points_tile(url=new_url)
        points_data = translate_points(points_data, coords)

        points_map[img_id] = points_data
        matches_map[img_id] = match_data

    annotation_scales = [base_coords[dim]["scale"] for dim in ("x", "y", "z", "t")]  # type: ignore
    annotation_units = ["um", "um", "um", "s"]
    annotation_space = neuroglancer.CoordinateSpace(
        names=["x", "y", "z", "t"], scales=annotation_scales, units=annotation_units
    )

    point_color = tile_coordinate_to_rgba(image_name_to_tile_coord(tile_name))

    line_starts: list[tuple[float, float, float, float]] = []
    line_stops: list[tuple[float, float, float, float]] = []
    point_map_self = points_map[image_id]

    # save points for self
    # pad with 0 for time coordinate
    point_data = [(*p, 0.0) for p in point_map_self.get_column("loc_xyz").to_list()]
    id_data = point_map_self.get_column("id").to_list()
    write_point_annotations(
        points_url,
        points=point_data,
        ids=id_data,
        coordinate_space=annotation_space,
        point_color=point_color,
    )
    if matches_map[image_id] is not None:
        log.info(f"Saving matches to {lines_url}.")
        match_entry = matches_map[image_id]
        if match_entry is None:
            raise ValueError(f"Missing match data for {image_id}")
        match_entry = cast(pl.DataFrame, match_entry)
        for row in match_entry.rows():
            point_self, point_other, id_self, id_other = row
            row_self = point_map_self.row(
                by_predicate=(pl.col("id") == point_self), named=True
            )
            line_start = (
                *row_self["loc_xyz"],
                0.0,
            )  # add a 0 for the time coordinate
            if len(line_start) != 4:
                msg = f"Wrong number of elements in line_start ({len(line_start)})"
                raise ValueError(msg)
            line_start = cast(tuple[float, float, float, float], line_start)
            try:
                row_other = points_map[id_other].row(
                    by_predicate=(pl.col("id") == point_other), named=True
                )
                line_stop = (
                    *row_other["loc_xyz"],
                    0.0,
                )  # add a 0 for the time coordinate
                if len(line_stop) != 4:
                    msg = f"Wrong number of elements in line_start ({len(line_start)})"
                line_stop = cast(tuple[float, float, float, float], line_stop)

            except pl.exceptions.NoRowsReturnedError:
                log.info(f"indexing error with {point_other} into vs_id {id_other}")
                line_stop = line_start

            line_starts.append(line_start)
            line_stops.append(line_stop)

        lines_loc = tuple(zip(*(line_starts, line_stops)))
        write_line_annotations(
            lines_url,
            lines=lines_loc,
            coordinate_space=annotation_space,
            point_color=point_color,
        )
    log.info(f"Completed saving points / matches after {time.time() - start:0.4f}s.")


def save_interest_points(*, bs_model: SpimData2, alignment_url: URL, dest: URL):
    """
    Save interest points for all tiles as collection of neuroglancer precomputed annotations. One
    collection of annotations will be generated per image described in the bigstitcher metadata under
    the directory name <out_prefix>/<image_name>.precomputed
    """

    view_setup_dict: dict[int, ViewSetup] = {
        int(v.ident): v for v in bs_model.sequence_description.view_setups.view_setups
    }

    # generate a coordinate grid for all the images
    tile_coords: dict[int, Coords] = get_tile_coords(bs_model=bs_model)
    if bs_model.view_interest_points is None:
        raise ValueError(
            "No view interest points were found in the bigstitcher xml file."
        )

    for file in bs_model.view_interest_points.data:
        setup_id = int(file.setup)
        tile_name = view_setup_dict[setup_id].name
        # todo: use pydantic zarr models to formalize this path
        fname = file.path.split("/")[0]
        _alignment_url = alignment_url.joinpath(f"interestpoints.n5/{fname}")
        save_annotations(
            image_id=setup_id,
            tile_name=tile_name,
            interest_points=_alignment_url,
            dest_url=dest,
            tile_coords=tile_coords,
        )


def fetch_summarize_matches(
    *, bigstitcher_xml: URL, pool: ThreadPoolExecutor, anon: bool = True
) -> pl.DataFrame:
    from structlog import get_logger

    log = get_logger()
    bs_model = read_bigstitcher_xml(bigstitcher_xml)
    interest_points_url = bigstitcher_xml.parent.joinpath("interestpoints.n5")
    all_matches = fetch_all_matches(
        bs_model=bs_model,
        n5_interest_points_url=interest_points_url,
        pool=pool,
        anon=anon,
    )
    if len(all_matches) == 0:
        raise ValueError("No matches found!")
    valid_matches = {}
    for key, value in all_matches.items():
        if isinstance(value, BaseException):
            log.info(f"An exception occurred when accessing {key}")
            log.exception(value)
        else:
            valid_matches[key] = value

    summarized = summarize_matches(bs_model=bs_model, matches_dict=valid_matches)
    return summarized


def summarize_match(match: pl.DataFrame) -> pl.DataFrame:
    """
    Summarize a dataframe of matches read from the bigstitcher n5 output by grouping by the id
    """
    return (
        match.group_by("id_self", "id_other")
        .agg(pl.col("point_self").count())
        .rename({"point_self": "num_matches"})
    )


def summarize_matches(
    bs_model: SpimData2, matches_dict: dict[str, pl.DataFrame]
) -> pl.DataFrame:
    tile_coords = get_tile_coords(bs_model)

    image_names = tuple(
        v.name for v in bs_model.sequence_description.view_setups.view_setups
    )

    # polars dataframes keyed by bigstitcher image names
    individual_summaries = {}
    for v in matches_dict.values():
        individual_summaries[v["id_self"][0]] = summarize_match(v)

    # include the file name and the normalized tile coordinate to each column
    individual_augmented = {}
    for k, v in individual_summaries.items():
        tcoord = tile_coords[k]["x"], tile_coords[k]["y"], tile_coords[k]["z"]
        individual_augmented[k] = v.with_columns(
            image_name=pl.lit(image_names[k]),
            x_coord_self=pl.lit(tcoord[0]),
            y_coord_self=pl.lit(tcoord[1]),
            z_coord_self=pl.lit(tcoord[2]),
        )
    result = (
        pl.concat(individual_augmented.values())
        .select(
            [
                pl.col("image_name"),
                pl.col("id_self"),
                pl.col("id_other"),
                pl.col("num_matches"),
                pl.col("x_coord_self"),
                pl.col("y_coord_self"),
                pl.col("z_coord_self"),
            ]
        )
        .sort("id_self")
    )
    return result


def fetch_all_matches(
    *,
    bs_model: SpimData2,
    n5_interest_points_url: URL,
    pool: ThreadPoolExecutor,
    anon: bool = True,
) -> dict[str, pl.DataFrame | BaseException | None]:
    """
    Load all the match data from the n5 datasets containing it.
    Takes a URL to an n5 group emitted by bigstitcher for storing interest points, e.g.
    s3://bucket/dataset/interestpoints.n5/.

    This function uses a thread pool to speed things up.
    """
    vips = bs_model.view_interest_points
    all_beads = tuple(n5_interest_points_url.joinpath(ip.path) for ip in vips.data)
    matches_dict: dict[str, pl.DataFrame | BaseException | None] = {}
    futures_dict = {}

    for bead_path in all_beads:
        fut = pool.submit(load_points_tile, bead_path)
        futures_dict[fut] = bead_path

    for result in as_completed(futures_dict):
        key = futures_dict[result]
        try:
            _, matches = result.result()
            matches_dict[key] = matches
        except BaseException:
            matches_dict[key] = result.exception()

    return matches_dict



from typing import Literal
from typing_extensions import TypedDict


class Tx(TypedDict):
    scale: float
    trans: float


class Coords(TypedDict):
    x: Tx
    y: Tx
    z: Tx


class TileCoordinate(TypedDict):
    x: int
    y: int
    z: int
    ch: Literal["488", "561"]



import pytest
import subprocess
import os
from click.testing import CliRunner
from matchviz.cli import tabulate_matches_cli


@pytest.mark.skip
@pytest.mark.parametrize(
    "alignment_url",
    [
        "s3://aind-open-data/exaSPIM_708373_2024-04-02_19-49-38_alignment_2024-05-07_18-15-25/bigstitcher.xml"
    ],
)
def test_save_points(tmpdir, alignment_url):
    out_path = os.path.join(str(tmpdir), "tile_alignment_visualization")
    run_result = subprocess.run(
        [
            "matchviz",
            "save-points",
            "--src",
            alignment_url,
            "--dest",
            out_path,
        ]
    )
    assert run_result.returncode == 0


@pytest.mark.skip
@pytest.mark.parametrize(
    "alignment_url",
    [
        "s3://aind-open-data/exaSPIM_708373_2024-04-02_19-49-38_alignment_2024-05-07_18-15-25/bigstitcher.xml"
    ],
)
def test_save_neuroglancer(tmpdir, alignment_url):
    points_url = os.path.join(alignment_url, "tile_alignment_visualization", "points")
    out_path = os.path.join(str(tmpdir), "tile_alignment_visualization", "neuroglancer")
    run_result = subprocess.run(
        [
            "matchviz",
            "ngjson",
            "--alignment-url",
            alignment_url,
            "--points-url",
            points_url,
            "--dest-path",
            out_path,
        ]
    )
    assert run_result.returncode == 0


@pytest.mark.skip
def test_summarize_points():
    alignment_url = "s3://aind-open-data/exaSPIM_708373_2024-04-02_19-49-38_alignment_2024-05-07_18-15-25/"
    runner = CliRunner()
    _ = runner.invoke(tabulate_matches_cli, ["--alignment-url", alignment_url])



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
import json
from typing import Any, Literal, Sequence
import click
import fsspec
import fsspec.implementations
import fsspec.implementations.local
from yarl import URL
from matchviz import (
    create_neuroglancer_state,
)
from matchviz.bigstitcher import (
    fetch_summarize_matches,
    get_tilegroup_url,
    read_bigstitcher_xml,
    save_interest_points,
)
from matchviz.core import parse_url
from matchviz.neuroglancer_styles import (
    NeuroglancerViewerStyle,
    neuroglancer_view_styles,
)
import structlog
from s3fs import S3FileSystem

from matchviz.plot import plot_matches


@click.group("matchviz")
def cli(): ...


log_level = click.option("--log-level", type=click.STRING, default="info")


@cli.command("plot-matches")
@click.option(
    "--bigstitcher-xml",
    type=click.STRING,
    help="URL pointing to a bigsticher xml document",
    required=True,
)
@click.option(
    "--dest",
    help="Name of a file in which to save the plot.",
    required=True,
    type=click.STRING,
)
@click.option("--invert-y-axis", type=click.BOOL, is_flag=True, default=False)
@click.option("--invert-x-axis", type=click.BOOL, is_flag=True, default=False)
def plot_matches_cli(
    bigstitcher_xml: str,
    dest: Any,
    invert_y_axis: bool,
    invert_x_axis: bool,
):
    pool = ThreadPoolExecutor(max_workers=16)
    anon = True

    bigstitcher_xml_normalized = parse_url(bigstitcher_xml)
    data = fetch_summarize_matches(
        bigstitcher_xml=bigstitcher_xml_normalized, pool=pool, anon=anon
    )

    fig = plot_matches(
        data=data,
        dataset_name=bigstitcher_xml_normalized.path,
        invert_x=invert_x_axis,
        invert_y=invert_y_axis,
    )
    fig.savefig(dest)


@cli.command("save-points")
@click.option("--src", type=click.STRING, required=True)
@click.option("--dest", type=click.STRING, required=True)
def save_interest_points_cli(src: str, dest: str):
    """
    Save bigstitcher interest points from n5 to neuroglancer precomputed annotations.
    """
    # strip trailing '/' from src and dest
    src_parsed = URL(src.rstrip("/"))
    dest_parsed = URL(dest.rstrip("/"))
    save_points(bigstitcher_url=src_parsed, dest=dest_parsed)


def save_points(bigstitcher_url: URL, dest: URL):
    bs_model = read_bigstitcher_xml(bigstitcher_url)
    save_interest_points(
        bs_model=bs_model, alignment_url=bigstitcher_url.parent, dest=dest
    )


@cli.command("ngjson")
@click.option("--bigstitcher-xml", type=click.STRING, required=True)
@click.option("--points-url", type=click.STRING, default=None)
@click.option("--matches-url", type=click.STRING, default=None)
@click.option("--dest-path", type=click.STRING, required=True)
@click.option("--style", type=click.STRING, multiple=True)
def save_neuroglancer_json_cli(
    bigstitcher_xml: str,
    dest_path: str,
    points_url: str | None,
    matches_url: str | None,
    style: Sequence[NeuroglancerViewerStyle] | None = None,
):
    """
    Generate a neuroglancer viewer state as a JSON document.
    """
    log = structlog.get_logger()
    bigstitcher_xml_url = URL(bigstitcher_xml)
    if points_url is not None:
        points_url_parsed = URL(points_url)
    else:
        points_url_parsed = None

    if matches_url is not None:
        matches_url_parsed = URL(matches_url)
    else:
        matches_url_parsed = None

    dest_path_parsed = dest_path.rstrip("/")
    if style is None or len(style) < 1:
        style = neuroglancer_view_styles
    for _style in style:
        out_path = save_neuroglancer_json(
            bigstitcher_xml=bigstitcher_xml_url,
            dest_url=dest_path_parsed,
            points_url=points_url_parsed,
            matches_url=matches_url_parsed,
            style=_style,
        )
        log.info(f"Saved neuroglancer JSON state for style {_style} to {out_path}")


def save_neuroglancer_json(
    *,
    bigstitcher_xml: str | URL,
    points_url: str | URL | None,
    matches_url: str | URL | None,
    dest_url: str | URL,
    style: NeuroglancerViewerStyle,
) -> URL:
    bs_xml_parsed = parse_url(bigstitcher_xml)
    points_url_parsed = parse_url(points_url)
    matches_url_parsed = parse_url(matches_url)
    dest_url_parsed = parse_url(dest_url)
    bs_model = read_bigstitcher_xml(bs_xml_parsed)
    tilegroup_s3_url = get_tilegroup_url(bs_model)
    state = create_neuroglancer_state(
        image_url=tilegroup_s3_url,
        points_url=points_url_parsed,
        matches_url=matches_url_parsed,
        style=style,
    )
    out_fname = f"{style}.json"
    out_path = dest_url_parsed.joinpath(out_fname)

    if dest_url_parsed.scheme == "s3":
        fs = S3FileSystem()
    else:
        fs = fsspec.implementations.local.LocalFileSystem(auto_mkdir=True)

    with fs.open(out_path, mode="w") as fh:
        fh.write(json.dumps(state.to_json(), indent=2))

    return out_path


@cli.command("tabulate-matches")
@click.option("--bigstitcher-xml", type=click.STRING, required=True)
@click.option("--interest-points", type=click.STRING, default=None)
@click.option("--output", type=click.STRING, default="csv")
def tabulate_matches_cli(bigstitcher_xml: str, output: Literal["csv"] | None):
    """
    Generate a tabular representation of the correspondence metadata generated by bigstitcher.
    """
    pool = ThreadPoolExecutor(max_workers=16)
    bigstitcher_xml_url = URL(bigstitcher_xml)
    anon = True
    summarized = fetch_summarize_matches(
        bigstitcher_xml=bigstitcher_xml_url, pool=pool, anon=anon
    )
    if output == "csv":
        click.echo(summarized.write_csv())
    else:
        raise ValueError(f'Format {output} is not recognized. Allowed values: ("csv",)')
